
from enum import Enum

import numpy as np
from filterpy.common import Q_discrete_white_noise
from filterpy.kalman import KalmanFilter
from scipy.linalg import block_diag

import motpy.metrics as metrics
from motpy.core import Box, Detection, Track, Vector
from motpy.metrics import angular_similarity, calculate_iou


""" The list of model presets below is not complete, more reasonable
options will be added in the future """


class ModelPreset(Enum):
    constant_velocity_and_static_box_size_2d = {'order_pos': 1, 'dim_pos': 2,
                                                'order_size': 0, 'dim_size': 2}

    constant_acceleration_and_static_box_size_2d = {'order_pos': 2, 'dim_pos': 2,
                                                    'order_size': 0, 'dim_size': 2}


def _base_dim_block(dt: float, order: int = 1):
    block = np.array([[1, dt, (dt**2) / 2],
                      [0, 1, dt],
                      [0, 0, 1]])
    cutoff = order + 1
    return block[:cutoff, :cutoff]


def _zero_pad(arr, length: int):
    ret = np.zeros((length,))
    ret[:arr.shape[0]] = arr
    return ret


class Model:
    def __init__(
            self,
            dt: float,
            order_pos: int = 1,
            dim_pos: int = 2,
            order_size: int = 0,
            dim_size: int = 2,
            q_var_pos: float = 70.0,
            q_var_size: float = 10.0,
            r_var_pos: float = 1,
            r_var_size: float = 1,
            p_cov_p0: float = 1000.):

        self.dt = dt
        self.order_pos = order_pos
        self.dim_pos = dim_pos
        self.order_size = order_size
        self.dim_size = dim_size

        self.q_var_pos = q_var_pos
        self.q_var_size = q_var_size
        self.r_var_pos = r_var_pos
        self.r_var_size = r_var_size
        self.p_cov_p0 = p_cov_p0

        if self.order_pos > 2 or self.order_size > 2:
            raise ValueError('Currently only system orders <= 2 are supported')

        # the expected input/output box length
        self.dim_box = 2 * max(self.dim_pos, self.dim_size)

        # precalculate utility indexes
        self.pos_idxs, self.size_idxs, self.z_in_x_idxs, self.offset_idx = self._calc_idxs()

        # number of variables in model state
        self.state_length = self.dim_pos * (self.order_pos + 1) + \
            self.dim_size * (self.order_size + 1)

        # length of z (observation) vector
        self.measurement_length = self.dim_pos + self.dim_size

    def _calc_idxs(self):
        offset_idx = max(self.dim_pos, self.dim_size)

        pos_idxs = [pidx * (self.order_pos + 1)
                    for pidx in range(self.dim_pos)]

        size_idxs = [self.dim_pos * (self.order_pos + 1) + sidx * (self.order_size + 1)
                     for sidx in range(self.dim_size)]

        # indexes of measured quantities (z) in state (x) vector
        z_in_x_idxs = np.concatenate((pos_idxs, size_idxs))

        return np.array(pos_idxs), np.array(size_idxs), z_in_x_idxs, offset_idx

    def build_F(self):
        """ returns constructed F matrix with specified positional
            e.g. (x,y,z) and size e.g. (w,h) dimensions """
        block_pos = _base_dim_block(self.dt, self.order_pos)
        block_size = _base_dim_block(self.dt, self.order_size)
        diag_components = [block_pos] * self.dim_pos + [block_size] * self.dim_size
        return block_diag(*diag_components)

    def build_Q(self):
        """ process noise """
        var_pos = self.q_var_pos
        var_size = self.q_var_size

        q_pos = var_pos if self.order_pos == 0 else Q_discrete_white_noise(
            dim=self.order_pos + 1, dt=self.dt, var=var_pos)

        q_size = var_size if self.order_size == 0 else Q_discrete_white_noise(
            dim=self.order_size + 1, dt=self.dt, var=var_size)

        diag_components = [q_pos] * self.dim_pos + [q_size] * self.dim_size
        return block_diag(*diag_components)

    def build_R(self):
        """ measurement noise, expected order is positon first, then size """
        block_pos = np.eye(self.dim_pos) * self.r_var_pos
        block_size = np.eye(self.dim_size) * self.r_var_size
        return block_diag(block_pos, block_size)

    def build_H(self):
        """ measurement matrix """
        # we only measure the first variable in each dimension
        def _base_block(order): return np.array([1] + [0] * order)
        diag_components = \
            [_base_block(self.order_pos)] * self.dim_pos +\
            [_base_block(self.order_size)] * self.dim_size
        return block_diag(*diag_components)

    def build_P(self):
        return np.eye(self.state_length) * self.p_cov_p0

    def box_to_z(self, box: Box) -> Vector:
        assert self.dim_box == len(box)
        box = np.array(box).reshape(2, (int(self.dim_box / 2)))
        center = (np.sum(box, axis=0) / 2.0)[:self.dim_pos]
        length = (box[1, :] - box[0, :])[:self.dim_size]
        return np.concatenate((center, length))

    def box_to_x(self, box: Box) -> Vector:
        """ box is expected to be in [xmin, ymin, zmin, ..., xmax, ymax, zmax, ...] format
        for 2d1ord+2d0ord case returns np.array([cx, 0, 0, cy, 0, 0, w, h]) """
        x = np.zeros((self.state_length,))
        x[self.z_in_x_idxs] = self.box_to_z(box)
        return x

    def x_to_box(self, x):
        size = max(self.dim_pos, self.dim_size)
        center = _zero_pad(x[self.pos_idxs], size)
        length = _zero_pad(x[self.size_idxs], size)
        return np.concatenate((center - length / 2, center + length / 2))



import uuid
from typing import Optional, Sequence, Union

import numpy as np
import scipy
from filterpy.kalman import KalmanFilter
from loguru import logger

from motpy.core import Box, Detection, Track, Vector
from motpy.metrics import angular_similarity, calculate_iou
from motpy.model import ModelPreset, Model


def get_object_tracker(dt: float, model: Model, x0: Optional[Vector] = None):
    """ returns Kalman-based tracker based on a specified motion model spec.
        e.g. for spec = {'order_pos': 1, 'dim_pos': 2, 'order_size': 0, 'dim_size': 1}
        we expect the following setup:
        state x, x', y, y', w, h
        where x and y are centers of boxes
              w and h are width and height
    """

    tracker = KalmanFilter(dim_x=model.state_length,
                           dim_z=model.measurement_length)
    tracker.F = model.build_F()
    tracker.Q = model.build_Q()
    tracker.H = model.build_H()
    tracker.R = model.build_R()
    tracker.P = model.build_P()

    if x0 is not None:
        tracker.x = x0

    return tracker


DEFAULT_MODEL_SPEC = ModelPreset.constant_velocity_and_static_box_size_2d.value


class Tracker:
    def __init__(
            self,
            model_spec: dict = DEFAULT_MODEL_SPEC,
            dt: float = 1 / 24,
            x0: Optional[Vector] = None,
            box0: Optional[Box] = None,
            max_staleness: float = 12.0):
        self.id = str(uuid.uuid4())
        self.status = 'unconfirmed'
        self.model_spec = model_spec

        self.steps_alive = 1
        self.steps_positive = 1
        self.staleness = 0.0
        self.max_staleness = max_staleness

        self.feature = None

        logger.debug(
            'creating new object tracker with %s and id %s' % (self.model_spec, self.id))

        self.model = Model(dt=dt, **self.model_spec)

        if x0 is None:
            x0 = self.model.box_to_x(box0)

        self._tracker = get_object_tracker(dt=dt, x0=x0, model=self.model)

    def predict(self):
        self.steps_alive += 1
        self._tracker.predict()

    def update(self, detection: Detection):
        self.steps_positive += 1

        # KF tracker update for position and size
        z = self.model.box_to_z(detection.box)
        self._tracker.update(z)

        if detection.feature is not None:
            self.update_feature(detection.feature)

        # reduce the staleness of a tracker, faster than growth rate
        self.unstale(rate=3)

    def update_feature(self, feature: Vector, alpha: float = 0.1):
        if self.feature is None:
            self.feature = np.array(feature)
        else:
            self.feature = alpha * np.array(feature) + (1 - alpha) * self.feature

    def stale(self, rate: float = 1.0):
        self.staleness += 1
        return self.staleness

    def unstale(self, rate: float = 2.0):
        self.staleness = max(0, self.staleness - rate)
        return self.staleness

    @property
    def is_stale(self):
        return self.staleness >= self.max_staleness

    @property
    def is_invalid(self):
        try:
            has_nans = any(np.isnan(self._tracker.x))
            return has_nans
        except Exception as e:
            logger.trace('invalid tracker, exception: %s' % str(e))
            return True

    @property
    def box(self):
        return self.model.x_to_box(self._tracker.x)

    def __repr__(self):
        fmt = "(box) %s\n (status) %s\n(staleness) %d"
        return fmt % (self.box, self.status, self.staleness)


""" assignment cost calculation & matching methods """


def match_by_cost_matrix(trackers: Sequence[Tracker],
                         detections: Sequence[Detection],
                         min_iou: float = 0.1,
                         **kwargs):
    if len(trackers) == 0 or len(detections) == 0:
        return []

    cost_mat, iou_mat = cost_matrix_iou_feature(trackers, detections, **kwargs)
    row_ind, col_ind = scipy.optimize.linear_sum_assignment(cost_mat)

    # filter out low IOU matches
    ret = [[r, c] for r, c in zip(row_ind, col_ind) if iou_mat[r, c] >= min_iou]
    return np.array(ret)


def _sequence_has_none(seq):
    return any(r is None for r in seq)


def cost_matrix_iou_feature(trackers: Sequence[Tracker],
                            detections: Sequence[Detection],
                            feature_similarity_fn=angular_similarity,
                            feature_similarity_beta: float = None):

    # boxes
    b1 = np.array([t.box for t in trackers])
    b2 = np.array([d.box for d in detections])

    # box iou
    inferred_dim = int(len(b1[0]) / 2)
    iou_mat = calculate_iou(b1, b2, dim=inferred_dim)

    # feature similarity
    if feature_similarity_beta is not None:
        # get features
        f1 = [t.feature for t in trackers]
        f2 = [d.feature for d in detections]

        if _sequence_has_none(f1) or _sequence_has_none(f2):
            # fallback to pure IOU due to missing features
            apt_mat = iou_mat
        else:
            sim_mat = feature_similarity_fn(f1, f2)
            sim_mat = feature_similarity_beta + (1 - feature_similarity_beta) * sim_mat

            # combined aptitude
            apt_mat = np.multiply(iou_mat, sim_mat)
    else:
        apt_mat = iou_mat

    cost_mat = -1.0 * apt_mat
    return cost_mat, iou_mat


class MatchingFunction:
    def __call__(self,
                 trackers: Sequence[Tracker],
                 detections: Sequence[Detection]) -> np.ndarray:
        raise NotImplementedError()


class BasicMatchingFunction(MatchingFunction):
    """ It implements the most basic matching function, taking
    detections boxes and optional feature similarity into account """

    def __init__(self, min_iou: float = 0.1,
                 feature_similarity_fn=angular_similarity,
                 feature_similarity_beta: Optional[float] = None) -> None:

        self.min_iou = min_iou
        self.feature_similarity_fn = feature_similarity_fn
        self.feature_similarity_beta = feature_similarity_beta

    def __call__(self,
                 trackers: Sequence[Tracker],
                 detections: Sequence[Detection]) -> np.ndarray:
        return match_by_cost_matrix(
            trackers, detections,
            self.min_iou,
            feature_similarity_fn=self.feature_similarity_fn,
            feature_similarity_beta=self.feature_similarity_beta)


class MultiObjectTracker:
    def __init__(self, dt: float,
                 model_spec: Union[str, dict] = DEFAULT_MODEL_SPEC,
                 matching_fn: Optional[MatchingFunction] = None,
                 tracker_kwargs: dict = None,
                 matching_fn_kwargs: dict = None,
                 active_tracks_kwargs: dict = None):
        """
            model_spec specifies the dimension and order for position and size of the object
            matching_fn determines the strategy on which the trackers and detections are assigned.

            tracker_kwargs are passed to each single object tracker
            active_tracks_kwargs limits surfacing of fresh/fading out tracks
        """

        self.dt = dt
        self.trackers = []

        if isinstance(model_spec, dict):
            self.model_spec = model_spec
        elif isinstance(model_spec, str) and model_spec in ModelPreset.__members__:
            self.model_spec = ModelPreset[model_spec].value
        else:
            raise NotImplementedError('unsupported motion model %s' % str(model_spec))
        logger.trace('using model spec: %s' % str(self.model_spec))

        self.matching_fn = matching_fn
        self.matching_fn_kwargs = matching_fn_kwargs if matching_fn_kwargs is not None else {}
        if self.matching_fn is None:
            self.matching_fn = BasicMatchingFunction(**self.matching_fn_kwargs)

        # kwargs to be passed to each single object tracker
        self.tracker_kwargs = tracker_kwargs if tracker_kwargs is not None else {}
        logger.trace('using tracker_kwargs: %s' % str(self.tracker_kwargs))

        # kwargs to be used when self.step returns active tracks
        self.active_tracks_kwargs = active_tracks_kwargs if active_tracks_kwargs is not None else {}
        logger.trace('using active_tracks_kwargs: %s' % str(self.active_tracks_kwargs))

    def active_tracks(self,
                      max_staleness_to_positive_ratio: float = 3.0,
                      max_staleness: float = 999,
                      min_steps_alive: int = -1):
        """ returns all active tracks after optional filtering by tracker steps count and staleness """

        tracks = []
        for tracker in self.trackers:
            cond1 = tracker.staleness / tracker.steps_positive < max_staleness_to_positive_ratio  # early stage
            cond2 = tracker.staleness < max_staleness
            cond3 = tracker.steps_alive >= min_steps_alive
            if cond1 and cond2 and cond3:
                tracks.append(Track(id=tracker.id, box=tracker.box))

        logger.trace('active/all tracks: %d/%d' % (len(self.trackers), len(tracks)))
        return tracks

    def cleanup_trackers(self):
        count_before = len(self.trackers)
        self.trackers = [t for t in self.trackers if not (t.is_stale or t.is_invalid)]
        count_after = len(self.trackers)
        logger.trace('deleted %s/%s trackers' % (count_before - count_after, count_before))

    def step(self, detections: Sequence[Detection]):
        """ the method matches the new detections with existing trackers,
        creates new trackers if necessary and performs the cleanup.
        Returns the active tracks after active filtering applied """

        # filter out empty detections
        detections = [det for det in detections if det.box is not None]

        logger.trace('step with %d detections' % len(detections))
        matches = self.matching_fn(self.trackers, detections)
        logger.trace('matched %d pairs' % len(matches))

        # all trackers: predict
        for t in self.trackers:
            t.predict()

        # assigned trackers: correct
        for match in matches:
            track_idx, det_idx = match[0], match[1]
            self.trackers[track_idx].update(detection=detections[det_idx])

        # not assigned detections: create new trackers POF
        assigned_det_idxs = set(matches[:, 1]) if len(matches) > 0 else []
        for det_idx in set(range(len(detections))).difference(assigned_det_idxs):
            tracker = Tracker(box0=detections[det_idx].box,
                              model_spec=self.model_spec,
                              **self.tracker_kwargs)
            self.trackers.append(tracker)

        # unassigned trackers
        assigned_track_idxs = set(matches[:, 0]) if len(matches) > 0 else []
        for track_idx in set(range(len(self.trackers))).difference(assigned_track_idxs):
            self.trackers[track_idx].stale()

        # cleanup dead trackers
        self.cleanup_trackers()

        return self.active_tracks(**self.active_tracks_kwargs)



import collections
import sys
from typing import Optional

import numpy as np
from loguru import logger

""" types """

# Box is of shape (1,2xdim), e.g. for dim=2 [xmin, ymin, xmax, ymax] format is accepted
Box = np.ndarray

# Vector is of shape (1, N)
Vector = np.ndarray

# Track is meant as an output from the object tracker
Track = collections.namedtuple('Track', 'id box')


class Detection:
    # Detection is to be an input the the tracker
    def __init__(
            self,
            box: Box,
            score: Optional[float] = None,
            feature: Optional[Vector] = None):
        self.box = box
        self.score = score
        self.feature = feature

    def __repr__(self):
        fmt = "(box) %s,\t(score) %s,\t(feature) %s\n"
        return fmt % (str(self.box),
                      str(self.score) or 'none',
                      str(self.feature) or 'none')


""" utils """


def set_log_level(level: str) -> None:
    logger.remove()
    logger.add(sys.stdout, level=level)



import numpy as np
from loguru import logger

from motpy.core import Track, Detection
from motpy.testing import CANVAS_SIZE, data_generator

try:
    import cv2
except BaseException:
    logger.error(
        'Could not import opencv. Please install opencv-python package or some of the testing functionalities will not be available')

""" methods below require opencv-python package installed """


def draw_rectangle(img, box, color, thickness: int = 3):
    img = cv2.rectangle(img, (int(box[0]), int(box[1])), (int(
        box[2]), int(box[3])), color, thickness)
    return img


def draw_text(img, text, above_box, color=(255, 255, 255)):
    tl_pt = (int(above_box[0]), int(above_box[1]) - 7)
    cv2.putText(img, text, tl_pt,
                fontFace=cv2.FONT_HERSHEY_SIMPLEX,
                fontScale=0.5, color=color)
    return img


def draw_track(img, track: Track, random_color: bool = True, fallback_color=(200, 20, 20)):
    color = [ord(c) * ord(c) % 256 for c in track.id[:3]] if random_color else fallback_color
    img = draw_rectangle(img, track.box, color=color, thickness=5)
    img = draw_text(img, track.id[:5] + '...', above_box=track.box)
    return img


def draw_detection(img, detection: Detection):
    img = draw_rectangle(img, detection.box, color=(0, 220, 0), thickness=1)
    return img


def image_generator(*args, **kwargs):

    def _empty_canvas(canvas_size=(CANVAS_SIZE, CANVAS_SIZE, 3)):
        img = np.ones(canvas_size, dtype=np.uint8) * 30
        return img

    data_gen = data_generator(*args, **kwargs)
    for dets_gt, dets_pred in data_gen:
        img = _empty_canvas()

        # overlay actor shapes
        for det_gt in dets_gt:
            xmin, ymin, xmax, ymax = det_gt.box
            feature = det_gt.feature
            for channel in range(3):
                img[int(ymin):int(ymax), int(xmin):int(xmax), channel] = feature[channel]

        yield img, dets_gt, dets_pred


if __name__ == "__main__":
    for img, dets_gt, dets_pred in image_generator(
            num_steps=1000, num_objects=10):

        for det_gt, det_pred in zip(dets_gt, dets_pred):
            img = draw_rectangle(img, det_gt.box, color=det_gt.feature)

            if det_pred.box is not None:
                img = draw_rectangle(img, det_pred.box, color=det_pred.feature, thickness=1)

        cv2.imshow('preview', img)
        c = cv2.waitKey(33)
        if c == ord('q'):
            break



import numpy as np
from scipy.spatial.distance import cdist


def calculate_iou(bboxes1, bboxes2, dim: int = 2):
    """ expected bboxes size: (-1, 2*dim) """
    bboxes1 = np.array(bboxes1).reshape((-1, dim * 2))
    bboxes2 = np.array(bboxes2).reshape((-1, dim * 2))

    coords_b1 = np.split(bboxes1, 2 * dim, axis=1)
    coords_b2 = np.split(bboxes2, 2 * dim, axis=1)

    coords_tl, coords_br = dict(), dict()  # tl and br points
    val_inter, val_b1, val_b2 = 1.0, 1.0, 1.0
    for d in range(dim):
        coords_tl[d] = np.maximum(coords_b1[d], np.transpose(coords_b2[d]))
        coords_br[d] = np.minimum(coords_b1[d + dim], np.transpose(coords_b2[d + dim]))

        val_inter *= np.maximum(coords_br[d] - coords_tl[d], 0)
        val_b1 *= coords_b1[d + dim] - coords_b1[d]
        val_b2 *= coords_b2[d + dim] - coords_b2[d]

    iou = val_inter / (val_b1 + np.transpose(val_b2) - val_inter)
    return iou


def angular_similarity(vectors1, vectors2):
    sim = 1 - cdist(vectors1, vectors2, 'cosine') / 2  # kept in range <0,1>
    return sim



import math
import random

import numpy as np

from motpy.core import Detection

CANVAS_SIZE = 1000


def _random_color():
    color_rgb = [random.randint(0, 255) for _ in range(3)]
    return color_rgb


class Actor():
    """ Actor is a box moving in 2d space """

    def __init__(self,
                 color=None,
                 max_omega: float = 0.05,
                 miss_prob: float = 0.1,
                 disappear_prob: float = 0.01,
                 det_err_sigma: float = 1.0,
                 canvas_size: int = 400):

        self.max_omega = max_omega
        self.miss_prob = miss_prob
        self.disappear_prob = disappear_prob
        self.det_err_sigma = det_err_sigma
        self.canvas_size = canvas_size

        # randomize size
        self.width = random.randint(50, 120)
        self.height = random.randint(50, 120)

        # randomize motion
        self.omega_x = random.uniform(-self.max_omega, self.max_omega)
        self.omega_y = random.uniform(-self.max_omega, self.max_omega)
        self.fi_x = random.randint(-180, 180)
        self.fi_y = random.randint(-90, 90)

        # let's treat color as a kind of feature
        if color is None:
            self.color = _random_color()

        self.disappear_steps = 0

    def position_at(self, step: int):
        half = self.canvas_size / 2 - 50
        x = half * math.cos(self.omega_x * step + self.fi_x) + half
        y = half * math.cos(self.omega_y * step + self.fi_y) + half
        return (x, y)

    def detections(self, step: int):
        """ returns ground truth and potentially missing detection for a given actor """
        xmin, ymin = self.position_at(step)
        box_gt = [xmin, ymin, xmin + self.width, ymin + self.height]

        # detection has some noise around the face coordinates
        box_pred = [random.gauss(0, self.det_err_sigma) + v for v in box_gt]

        # due to flakyness, some detections are missing
        if random.random() < self.miss_prob:
            box_pred = None

        # disappear_steps
        if random.random() < self.disappear_prob:
            self.disappear_steps = random.randint(1, 24)

        if self.disappear_steps > 0:
            box_pred = None
            self.disappear_steps -= 1

        # wrap boxes and features as detections
        det_gt = Detection(box=box_gt, score=1.0, feature=self.color)

        feature_pred = [random.gauss(0, 5) + v for v in self.color]
        det_pred = Detection(box=box_pred,
                             score=random.random() / 2 + 0.5,
                             feature=feature_pred)

        return det_gt, det_pred


def data_generator(
        num_steps: int = 1000,
        num_objects: int = 1,
        max_omega: float = 0.01,
        miss_prob: float = 0.1,
        disappear_prob: float = 0.0,
        det_err_sigma: float = 1.0):

    actors = [Actor(max_omega=max_omega,
                    miss_prob=miss_prob,
                    disappear_prob=disappear_prob,
                    det_err_sigma=det_err_sigma,
                    canvas_size=CANVAS_SIZE) for _ in range(num_objects)]

    for step in range(num_steps):
        dets_gt, dets_pred = [], []

        for actor in actors:
            det_gt, det_pred = actor.detections(step)
            dets_gt.append(det_gt)
            dets_pred.append(det_pred)

        yield dets_gt, dets_pred



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='motpy',
    version='0.0.8',
    url='https://github.com/wmuron/motpy.git',
    download_url='https://github.com/wmuron/motpy/releases/tag/v0.0.7-alpha',
    author='Wiktor Muron',
    author_email='wiktormuron@gmail.com',
    description='Library for track-by-detection multi object tracking implemented in python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(exclude=("tests",)),
    python_requires='>3.6',
    install_requires=['numpy',
                      'scipy',
                      'filterpy',
                      'loguru'],
    license='MIT',
    keywords=['multi-object-tracking', 'object-tracking', 'kalman-filter'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License'
    ]
)
